BootStrap: library
From: debian

%post
    apt update
    apt -y --allow-unauthenticated full-upgrade
    apt -y --allow-unauthenticated install --no-install-recommends git python3 python3-pip
    apt clean
    pip install "kraken[pdf,augment]"
    pip install "numpy<1.24"
    pip cache purge
    mkdir /sps /pbs
